/* -*- buffer-read-only: t -*- vi: set ro:
   THIS FILE IS GENERATED AUTOMATICALLY.  PLEASE DO NOT EDIT.
*/
#define MF_SOURCE_NAME  MF_SOURCE_BI_IO
#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include "mailfromd.h"
#include "prog.h"
#include "builtin.h"

#line 1 "bi_io.m4"
/* This file is part of Mailfromd.             -*- c -*-
   Copyright (C) 2006, 2007, 2008, 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#define NSTREAMS 1024

struct io_stream {
	char *name;
	mu_locker_t lock;
	int fd[2];
	pid_t pid;
	char *buf;
	size_t bufsize;
	void (*cleanup)(void*);
	void *cleanup_data;
};

#define IFD(s) ((s).fd[0])
#define OFD(s) ((s).fd[1] == -1 ? (s).fd[0] : (s).fd[1])

static void
flush_stream(struct io_stream *str)
{
	/*FIXME*/
}

static void
close_stream(struct io_stream *str)
{
	if (OFD(*str) == -1)
		return;
	flush_stream(str);
	close(OFD(*str));
	if (IFD(*str) != -1)
		close(IFD(*str));
	if (str->pid) {
		int status;
		waitpid(str->pid, &status, 0);
	}
	str->fd[0] = -1;
	str->fd[1] = -1;
	str->pid = 0;
	if (str->cleanup)
		str->cleanup(str->cleanup_data);
	str->cleanup = NULL;
	str->cleanup_data = NULL;
}

/* Read bytes from the stream STR into its buffer, until
   DELIM is encountered. Return number of bytes read. */
static int
read_stream_delim(struct io_stream *str, int delim)
{
	int fd = IFD(*str);
	size_t i = 0;
	int rc;

	for (i = 0; ; i++) {
		if (str->bufsize == i) {
			if (str->bufsize == 0) 
				str->bufsize = 16;
			str->buf = x2nrealloc(str->buf, &str->bufsize,
					      sizeof str->buf[1]);
		}
		rc = read(fd, str->buf + i, 1);
		if (rc == -1)
			return -1;
		else if (rc == 0)
			return 0;
		if ((unsigned char) str->buf[i] == delim) {
			/* NOTE: We're leaving the loop, so
			   we must increment the counter here: */
			str->buf[i++] = 0;
			break;
		}
	}
	return i;
}

#define REDIRECT_STDIN_P(f) ((f) & (O_WRONLY|O_RDWR))
#define REDIRECT_STDOUT_P(f) ((f) == O_RDONLY || ((f) & (O_RDONLY|O_RDWR)))

static int
open_program_stream(eval_environ_t env,
		    struct io_stream *str, const char *cmdline, int flags)
{
	int rightp[2], leftp[2];
	int rc = 0;
	pid_t pid;
	int argc;
	char **argv;
	
	if (REDIRECT_STDIN_P(flags))
		pipe(leftp);
	if (REDIRECT_STDOUT_P(flags))
		pipe(rightp);

	switch (pid = fork()) {
		/* The child branch.  */
	case 0:
		/* attach the pipes */

		/* Right-end */
		if (REDIRECT_STDOUT_P(flags)) {
			if (rightp[1] != 1) {
				close(1);
				dup2(rightp[1], 1);
			}
			close(rightp[0]);
		}

		/* Left-end */
		if (REDIRECT_STDIN_P(flags)) {
			if (leftp[0] != 0) {
				close(0);
				dup2(leftp[0], 0);
			}
			close(leftp[1]);
		}
      
		/* FIXME: Error output */
#if 0
		if (errfile) {
			i = open (errfile, O_CREAT|O_WRONLY|O_APPEND, 0644);
			if (i > 0 && i != 2) {
				dup2(i, 2);
				close(i);
			}
		}
#endif
		/* Close unneded descripitors */
		close_fds_above(2);

		debug1(10, "running %s", cmdline);
		rc = mu_argcv_get(cmdline, "", NULL, &argc, &argv);
		if (rc) {
			mu_error(_("cannot parse command line %s: %s"),
				 cmdline, mu_strerror(rc));
			exit(127);
		}			
		execvp(argv[0], argv);
		mu_error(_("cannot run %s: %s"),
			 cmdline, mu_strerror(rc));
		exit(127);
		/********************/

		/* Parent branches: */
	case -1:
		/* Fork has failed */
		/* Restore things */
		rc = errno;
		if (REDIRECT_STDOUT_P(flags)) {
			close(rightp[0]);
			close(rightp[1]);
		}
		if (REDIRECT_STDIN_P(flags)) {
			close(leftp[0]);
			close(leftp[1]);
		}
		break;
		
	default:
		str->pid = pid;
		if (REDIRECT_STDOUT_P(flags)) {
			str->fd[0] = rightp[0];
			close(rightp[1]);
		} else
			str->fd[0] = -1;

		if (REDIRECT_STDIN_P(flags)) {
			str->fd[1] = leftp[1];
			close(leftp[0]);
		} else
			str->fd[1] = -1;
	}
	return rc;
}


static int
open_file_stream(eval_environ_t env,
		 struct io_stream *str, const char *file, int flags)
{
	str->fd[0] = open(file, flags, 0644); /* FIXME: mode? */
	if (str->fd[0] == -1)
		return errno;
	return 0;
}



static int
open_parsed_inet_stream(eval_environ_t env,
			struct io_stream *str,
			const char *cstr,
			char *proto, char *port, char *path,
			int flags)
{
	struct timeval start, connect_tv;
	union {
		struct sockaddr sa;
		struct sockaddr_in s_in;
		struct sockaddr_un s_un;
	} addr;

	int socklen;
	int fd;

	if (!proto
	    || strcmp(proto, "unix") == 0 || strcmp(proto, "local") == 0) {
		struct stat st;
		
			if (!(port == NULL))
#line 229
			env_throw_bi(env, mfe_failure, NULL, _("invalid connection type: %s; "
			    "port is meaningless for UNIX sockets"),cstr)
#line 233
;
		
			if (!(strlen(path) <= sizeof addr.s_un.sun_path))
#line 235
			env_throw_bi(env, mfe_range, NULL, _("%s: UNIX socket name too long"),path)
#line 238
;
		
		addr.sa.sa_family = PF_UNIX;
		socklen = sizeof(addr.s_un);
		strcpy(addr.s_un.sun_path, path);
		
		if (stat(path, &st)) {
				env_throw_bi(env, mfe_failure, NULL, _("%s: cannot stat socket: %s"),path,strerror(errno));
#line 248
		} else {
			/* FIXME: Check permissions? */
				if (!(S_ISSOCK(st.st_mode)))
#line 250
			env_throw_bi(env, mfe_failure, NULL, _("%s: not a socket"),path)
#line 253
;
		}

	} else if (strcmp(proto, "inet") == 0) {
		short pnum;
		long num;
		char *p;
		
		addr.sa.sa_family = PF_INET;
		socklen = sizeof(addr.s_in);

			if (!(port != NULL))
#line 264
			env_throw_bi(env, mfe_failure, NULL, _("invalid connection type: %s; "
			    "missing port number"),cstr)
#line 268
;

		num = pnum = strtol(port, &p, 0);
		if (*p == 0) {
				if (!(num == pnum))
#line 272
			env_throw_bi(env, mfe_range, NULL, _("invalid connection type: "
				    "%s; bad port number"),cstr)
#line 276
;
			pnum = htons(pnum);
		} else {
			struct servent *sp = getservbyname(port, "tcp");

				if (!(sp != NULL))
#line 281
			env_throw_bi(env, mfe_failure, NULL, _("invalid connection type: "
				    "%s; unknown port name"),cstr)
#line 285
;
			pnum = sp->s_port;
		}
		
		if (!path)
			addr.s_in.sin_addr.s_addr = INADDR_ANY;
		else {
			struct hostent *hp = gethostbyname(path);
				if (!(hp != NULL))
#line 293
			env_throw_bi(env, mfe_failure, NULL, _("unknown host name %s"),path)
#line 296
;
			addr.sa.sa_family = hp->h_addrtype;
			switch (hp->h_addrtype) {
			case AF_INET:
				memmove(&addr.s_in.sin_addr, hp->h_addr, 4);
				addr.s_in.sin_port = pnum;
				break;

			default:
					env_throw_bi(env, mfe_range, NULL, _("invalid connection type: "
					   "%s; unsupported address family"),cstr);
#line 309
			}
		}
	} else {
			env_throw_bi(env, mfe_range, NULL, _("unsupported protocol: %s"),proto);
#line 315
	}

	fd = socket(addr.sa.sa_family, SOCK_STREAM, 0);
		if (!(fd != -1))
#line 318
			env_throw_bi(env, mfe_failure, NULL, _("unable to create new socket: %s"),strerror(errno))
#line 321
;

	/* FIXME: Bind to the source ? */

	connect_tv.tv_sec = 5;
	connect_tv.tv_usec = 0;
	
	gettimeofday(&start, NULL);
	for (;;) {
		int rc = connect(fd, &addr.sa, socklen);
		if (rc == 0)
			break;
		else if (errno == ECONNREFUSED) {
			struct timeval tv, now;
			fd_set rset, wset, xset;
			
			gettimeofday(&now, NULL);
			timersub(&now, &start, &tv);
			
			if (timercmp(&tv, &connect_tv, < )) {
				FD_ZERO(&rset);
				FD_SET(fd, &rset);
				FD_ZERO(&wset);
				FD_SET(fd, &wset);
				FD_ZERO(&xset);
				FD_SET(fd, &xset);
				select(fd + 1, &rset, &wset, &xset, &tv);
				continue;
			}
		}
		
		close(fd);
			env_throw_bi(env, mfe_failure, NULL, _("cannot connect to %s: %s"),cstr,strerror(errno));
#line 356
	}
	
	str->fd[0] = fd;
	return 0;
}

static int
open_inet_stream(eval_environ_t env,
		 struct io_stream *str, const char *addr, int flags)
{
	int rc;
	char *proto, *port, *path;

	if (gacopyz_parse_connection(addr, &proto, &port, &path)
	    != MI_SUCCESS) 
		rc = ENOMEM; /* FIXME: or EINVAL? */
	else {
		rc = open_parsed_inet_stream(env,
					     str, addr,
					     proto, port, path, flags);
		free(proto);
		free(port);
		free(path);
	}
	return rc;
}


static void *
alloc_streams()
{
	struct io_stream *p, *stab = xcalloc(NSTREAMS, sizeof *stab);
	for (p = stab; p < stab + NSTREAMS; p++) 
		p->fd[0] = p->fd[1] = -1;
	return stab;
}

static void
destroy_streams(void *data)
{
	struct io_stream *stab = data;
	struct io_stream *p;
	for (p = stab; p < stab + NSTREAMS; p++) {
		close_stream(p);
		free(p->buf);
	}
	free(stab);
}
		

#line 405

#line 405
static int IO_id;
#line 405 "bi_io.m4"



void
#line 408
bi_open(eval_environ_t env)
#line 408

#line 408 "bi_io.m4"
{
#line 408
	
#line 408

#line 408

#line 408
char * name;
#line 408
        
#line 408
get_string_arg(env, 0, &name);
#line 408
        
#line 408

#line 408
        adjust_stack(env, 1);
#line 408

#line 408
	if (prog_trace_option & BUILTIN_MASK_io)
#line 408
		prog_trace(env, "open %s",name);;
#line 408

{
	int i, rc;
	int flags = 0;
	int (*opf)(eval_environ_t env,
		   struct io_stream *, const char *, int) = open_file_stream;
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	
	for (i = 0; i < NSTREAMS; i++) {
		if (iotab[i].fd[0] == -1) 
			break;
	}
		if (!(i < NSTREAMS))
#line 420
			env_throw_bi(env, mfe_failure, "open", _("no more files available"))
#line 422
;

	debug1(10, "opening stream %s", name);
	iotab[i].name = name;
	if (*name == '>') {
		flags |= O_RDWR|O_CREAT;
		name++;
		if (*name == '>') {
			flags |= O_APPEND;
			name++;
		} else
			flags |= O_TRUNC;
	} else if (*name == '|') {
		opf = open_program_stream;
		flags = O_WRONLY;
		name++;
		if (*name == '&') {
			flags = O_RDWR;
			name++;
		}
	} else if (*name == '@') {
		name++;
		opf = open_inet_stream;
		flags = O_RDWR;
	} else
		flags = O_RDONLY;
	
	for (;*name && c_isspace(*name); name++)
		;
	
	rc = opf(env, &iotab[i], name, flags);
	
		if (!(rc == 0))
#line 454
			env_throw_bi(env, mfe_failure, "open", _("cannot open stream %s: %s"),name,mu_strerror(rc))
#line 457
;
	debug2(10, "open(%s) = %d", name, i);
	
#line 459
do {
#line 459
  push(env, (STKVAL) (i));
#line 459
  return;
#line 459
} while (0);
}
}

void
#line 463
bi_close(eval_environ_t env)
#line 463

#line 463 "bi_io.m4"
{
#line 463
	
#line 463

#line 463

#line 463
long  fd;
#line 463
        
#line 463
get_numeric_arg(env, 0, &fd);
#line 463
        
#line 463

#line 463
        adjust_stack(env, 1);
#line 463

#line 463
	if (prog_trace_option & BUILTIN_MASK_io)
#line 463
		prog_trace(env, "close %lu",fd);;
#line 463

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);

		if (!(fd >= 0 && fd < NSTREAMS))
#line 467
			env_throw_bi(env, mfe_range, "close", _("invalid file descriptor"))
#line 469
;
	close_stream(&iotab[fd]);
}
}	

void
#line 474
bi_write(eval_environ_t env)
#line 474

#line 474 "bi_io.m4"
{
#line 474
	
#line 474

#line 474
long __bi_argcnt;
#line 474
long  fd;
#line 474
        char * str;
#line 474
        long  n;
#line 474
        
#line 474
get_numeric_arg(env, 1, &fd);
#line 474
        get_string_arg(env, 2, &str);
#line 474
        get_numeric_arg(env, 3, &n);
#line 474
        
#line 474
get_numeric_arg(env, 0, &__bi_argcnt);
#line 474
        adjust_stack(env, __bi_argcnt + 1);
#line 474

#line 474
	if (prog_trace_option & BUILTIN_MASK_io)
#line 474
		prog_trace(env, "write %lu %s %lu",fd, str, ((__bi_argcnt > 2) ? n : 0));;
#line 474

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;
	
	debug2(10, "writing %s to %d", str, fd);
		if (!(fd >= 0 && fd < NSTREAMS && OFD(iotab[fd])))
#line 480
			env_throw_bi(env, mfe_range, "write", _("invalid file descriptor"))
#line 482
;
	if (!(__bi_argcnt > 2))
		n = strlen (str);
	rc = write(OFD(iotab[fd]), str, n);
		if (!(n == rc))
#line 486
			env_throw_bi(env, mfe_io, "write", _("write error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 489
;
}
}

void
#line 493
bi_read(eval_environ_t env)
#line 493

#line 493 "bi_io.m4"
{
#line 493
	
#line 493

#line 493

#line 493
long  fd;
#line 493
        long  size;
#line 493
        
#line 493
get_numeric_arg(env, 0, &fd);
#line 493
        get_numeric_arg(env, 1, &size);
#line 493
        
#line 493

#line 493
        adjust_stack(env, 2);
#line 493

#line 493
	if (prog_trace_option & BUILTIN_MASK_io)
#line 493
		prog_trace(env, "read %lu %lu",fd, size);;
#line 493

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;
	size_t off;
	char *s = (char*) env_data_ref(env, (off = heap_reserve(env, size + 1)));

		if (!(fd >= 0 && fd < NSTREAMS && IFD(iotab[fd])))
#line 500
			env_throw_bi(env, mfe_range, "read", _("invalid file descriptor"))
#line 502
;
	
	rc = read(IFD(iotab[fd]), s, size);
	if (rc == 0)
			env_throw_bi(env, mfe_eof, "read", _("EOF on %s"),iotab[fd].name);
#line 508
		if (!(rc == size))
#line 508
			env_throw_bi(env, mfe_io, "read", _("read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 511
;
	s[size] = 0;
	
#line 513
do {
#line 513
  push(env, (STKVAL) (off));
#line 513
  return;
#line 513
} while (0);
}	
}

/* FIXME: only first char of delim is used */
void
#line 518
bi_getdelim(eval_environ_t env)
#line 518

#line 518 "bi_io.m4"
{
#line 518
	
#line 518

#line 518

#line 518
long  fd;
#line 518
        char * delim;
#line 518
        
#line 518
get_numeric_arg(env, 0, &fd);
#line 518
        get_string_arg(env, 1, &delim);
#line 518
        
#line 518

#line 518
        adjust_stack(env, 2);
#line 518

#line 518
	if (prog_trace_option & BUILTIN_MASK_io)
#line 518
		prog_trace(env, "getdelim %lu %s",fd, delim);;
#line 518

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;

		if (!(fd >= 0 && fd < NSTREAMS && IFD(iotab[fd])))
#line 523
			env_throw_bi(env, mfe_range, "getdelim", _("invalid file descriptor"))
#line 525
;
	rc = read_stream_delim(&iotab[fd], delim[0]); 
	if (rc == 0)
			env_throw_bi(env, mfe_eof, "getdelim", _("EOF on %s"),iotab[fd].name);
#line 530
		if (!(rc > 0))
#line 530
			env_throw_bi(env, mfe_io, "getdelim", _("read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 533
;
	do {pushs(env, iotab[fd].buf);
#line 534
return;
#line 534
} while (0);
}	
}

void
#line 538
bi_getline(eval_environ_t env)
#line 538

#line 538 "bi_io.m4"
{
#line 538
	
#line 538

#line 538

#line 538
long  fd;
#line 538
        
#line 538
get_numeric_arg(env, 0, &fd);
#line 538
        
#line 538

#line 538
        adjust_stack(env, 1);
#line 538

#line 538
	if (prog_trace_option & BUILTIN_MASK_io)
#line 538
		prog_trace(env, "getline %lu",fd);;
#line 538

{
	struct io_stream *iotab = env_get_builtin_priv(env,IO_id);
	int rc;

		if (!(fd >= 0 && fd < NSTREAMS && IFD(iotab[fd])))
#line 543
			env_throw_bi(env, mfe_range, "getline", _("invalid file descriptor"))
#line 545
;
	rc = read_stream_delim(&iotab[fd], '\n');
	if (rc == 0)
			env_throw_bi(env, mfe_eof, "getline", _("EOF on %s"),iotab[fd].name);
#line 550
		if (!(rc > 0))
#line 550
			env_throw_bi(env, mfe_io, "getline", _("read error on %s: %s"),iotab[fd].name,mu_strerror(errno))
#line 553
;
	do {pushs(env, iotab[fd].buf);
#line 554
return;
#line 554
} while (0);
}	
}


#line 558
void
#line 558
io_init_builtin()
#line 558
{
#line 558
	#line 405 "bi_io.m4"
IO_id = builtin_priv_register(alloc_streams, destroy_streams,
#line 405
NULL);
#line 408 "bi_io.m4"
va_builtin_install_ex("open", bi_open, 0, dtype_number, 1, 0, 0|0, dtype_string);
#line 463 "bi_io.m4"
va_builtin_install_ex("close", bi_close, 0, dtype_unspecified, 1, 0, 0|0, dtype_number);
#line 474 "bi_io.m4"
va_builtin_install_ex("write", bi_write, 0, dtype_unspecified, 3, 1, 0|0, dtype_number, dtype_string, dtype_number);
#line 493 "bi_io.m4"
va_builtin_install_ex("read", bi_read, 0, dtype_string, 2, 0, 0|0, dtype_number, dtype_number);
#line 518 "bi_io.m4"
va_builtin_install_ex("getdelim", bi_getdelim, 0, dtype_string, 2, 0, 0|0, dtype_number, dtype_string);
#line 538 "bi_io.m4"
va_builtin_install_ex("getline", bi_getline, 0, dtype_string, 1, 0, 0|0, dtype_number);

#line 558 "bi_io.m4"
	
#line 558
}
#line 558

#line 627 "../mfd/snarf.m4"

#line 627

